/* usi.cc
 */
#include "osl/record/usi.h"
#include "osl/record/psn.h"
#include "osl/state/simpleState.h"
#include "osl/apply_move/applyMove.h"
#include <boost/algorithm/string/replace.hpp>
#include <boost/foreach.hpp>
#include <iostream>
#include <sstream>
#include <cctype>

const std::string osl::record::usi::
show(Move m)
{
  if (m.isPass())
    return "pass";
  if (! m.isNormal())
    return "resign";
  if (m == Move::DeclareWin())
    return "win";
  return psn::show(m);
}

const osl::Move osl::record::usi::
strToMove(const std::string& str, const SimpleState& s)
{
  return psn::strToMove(str, s);
}

osl::PtypeO osl::record::usi::
charToPtypeO(char c)
{
  const Ptype ptype = psn::charToPtype(toupper(c));
  if (ptype == PTYPE_EMPTY)
    throw ParseError("Invalid piece character: " + c);
  const Player pl = isupper(c) ? BLACK : WHITE;
  return newPtypeO(pl, ptype);
}

void osl::record::usi::parseBoard(const std::string& word, SimpleState& state)
{
  if (word.empty())
    throw ParseError(word);

  state.init();
  int x=9, y=1;
  for (size_t i=0; i<word.size(); ++i) {
    const char c = word[i];
    if (isalpha(c)) {
      const PtypeO ptypeo = charToPtypeO(c);
      state.setPiece(getOwner(ptypeo), Position(x,y), getPtype(ptypeo));
      --x;
    } else if (c == '+') {
      if ( (i+1) >= word.size() )
        throw ParseError(word);
      const char next = word[i+1];
      if (!isalpha(next))
        throw ParseError(word);
      const PtypeO ptypeo = charToPtypeO(next);
      if (!canPromote(ptypeo))
        throw ParseError(word);
      const PtypeO promoted = promote(ptypeo);
      state.setPiece(getOwner(promoted), Position(x,y), getPtype(promoted));
      --x;
      ++i;
    } else if (c == '/') {
      if (x != 0)
        throw ParseError(word);
      x = 9;
      ++y;
    } else if (isdigit(c)) {
      const int n = c - '0';
      if (n == 0)
        throw ParseError(word);
      x -= n;
    } else {
      throw ParseError("usi: unknown input " + c);
    }
    if (x < 0 || x > 9 || y < 0 || y > 9)
      throw ParseError(word);
  }
}

void osl::record::usi::parse(const std::string& line, SimpleState& state)
{
  vector<Move> moves;
  parse(line, state, moves);
  BOOST_FOREACH(Move move, moves) {
    ApplyMoveOfTurn::doMove(state, move);
  }
}

void osl::record::usi::parse(const std::string& line, SimpleState& state, vector<Move>& moves)
{
  moves.clear();
  std::istringstream is(line);
  std::string word;
  is >> word;
  if (word == "startpos") 
    state.init(HIRATE);
  else {
    if (word != "sfen")
        throw ParseError(word);
    is >> word;
    parseBoard(word, state);
    is >> word;
    if (word != "b" && word != "w")
        throw ParseError(word);
    state.setTurn((word == "b") ? BLACK : WHITE);
    is >> word;
    if (word != "-") {
      int prefix = 0;
      BOOST_FOREACH(char c, word) {
	if (isalpha(c)) {
	  PtypeO ptypeo = charToPtypeO(c);
	  for (int j=0; j<std::max(1, prefix); ++j)
	    state.setPiece(getOwner(ptypeo), Position::STAND(), getPtype(ptypeo));
	  prefix = 0;
	}
	else {
	  if (!isdigit(c))
            throw ParseError(word);
	  prefix = (c - '0') + prefix*10;
          if (prefix == 0)
            throw ParseError(word);
	}
      }
    }
    state.initPawnMask();
    int move_number; // will not be used
    if (! (is >> move_number))
      return;
    assert(is);
  }
  if (! (is >> word))
    return;
  if (word != "moves")
    throw ParseError(word);
  SimpleState state_copy = state;
  while (is >> word) {
    Move m = strToMove(word, state_copy);
    moves.push_back(m);
    ApplyMoveOfTurn::doMove(state_copy, m);
  }
} 

void osl::record::usi::
escape(std::string& str)
{
  boost::algorithm::replace_all(str, "/", "_");
  boost::algorithm::replace_all(str, "+", "@");
  boost::algorithm::replace_all(str, " ", ".");
}

void osl::record::usi::
unescape(std::string& str)
{
  boost::algorithm::replace_all(str, "_", "/");
  boost::algorithm::replace_all(str, "@", "+");
  boost::algorithm::replace_all(str, ".", " ");
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
